//---------------------------------------------------------------------------
#include <vcl.h>
#include <assert.h>
#include <registry.hpp>
#pragma hdrstop

#include "Mixmaster2.h"

// TODO 0: needs to do stats headers val checks with ignore case

TMixmaster3 *Mixmaster;
HINSTANCE hMIXLib = NULL;
HINSTANCE hSSLLib = NULL;
//---------------------------------------------------------------------------
// Utility
//---------------------------------------------------------------------------

bool __fastcall InitMixmaster(TComponent *Owner)
{
   Mixmaster = new TMixmaster3(Owner);

   if (Mixmaster->FailedInit)
   {
      delete Mixmaster;
      Mixmaster = NULL;
   }

   return Mixmaster != NULL;
}
//---------------------------------------------------------------------------

void __fastcall ShutDownMix()
{
   if (Mixmaster) delete Mixmaster;
}
//---------------------------------------------------------------------------

void __fastcall ShutDownMixmaster()
{
   if (Mixmaster) delete Mixmaster;
}
//---------------------------------------------------------------------------

AnsiString __fastcall MixDir()
{
   AnsiString dir = "";
   if (HasMix()) dir = Mixmaster->Directory;
   return dir;
}
//---------------------------------------------------------------------------

bool __fastcall HasMix()
{
   return Mixmaster != NULL;
}

//---------------------------------------------------------------------------
// TMixmaster3
//---------------------------------------------------------------------------

__fastcall TMixmaster3::TMixmaster3(TComponent *Owner)
   : TComponent(Owner)
{
   TRegistry *reg = new TRegistry();
   if (!reg->OpenKey("Software\\Mixmaster", false))
   {
      ShowMessage(
         "Mixmaster is not installed on this computer! Without it, I cannot continue!"
      );
      FFailedInit = true;
      delete reg;
      return;
   }

   FDirectory = reg->ReadString("MixDir");
   if (FDirectory.Length() && FDirectory.SubString(FDirectory.Length(), 1) != '\\')
      FDirectory += '\\';
   delete reg;

   hSSLLib = LoadLibrary(AnsiString(FDirectory + "libeay32.dll").c_str());
   hMIXLib = LoadLibrary(AnsiString(FDirectory + "mixlib.dll").c_str());
   // DONE 5: needs mix_init(Mixmaster->Directory) on startup

   FFailedInit = hSSLLib == NULL || hMIXLib == NULL;
   if (FFailedInit)
   {
      ShowMessage(
         "I was unable to initialize Mixmaster. Without it, I cannot continue!"
      );
   }
}
//---------------------------------------------------------------------------

__fastcall TMixmaster3::TMixmaster3(TComponent *Owner, const AnsiString &MixDir)
   : TComponent(Owner)
{
   FDirectory = MixDir;
   hSSLLib = LoadLibrary(AnsiString(FDirectory + "libeay32.dll").c_str());
   hMIXLib = LoadLibrary(AnsiString(FDirectory + "mixlib.dll").c_str());
   // DONE 5: needs mix_init(Mixmaster->Directory) on startup

   FFailedInit = hSSLLib == NULL || hMIXLib == NULL;
}
//---------------------------------------------------------------------------

__fastcall TMixmaster3::~TMixmaster3()
{
   // DONE 5: needs mixmaster cleanup on shutdown
   if (hMIXLib != NULL) FreeLibrary(hMIXLib);
   if (hSSLLib != NULL) FreeLibrary(hSSLLib);
}

//---------------------------------------------------------------------------
// MixLib Stuff
//---------------------------------------------------------------------------

int __fastcall TMixmaster3::mix_init(char *mixdir)
{
   MIX_INIT mixinit = (MIX_INIT)GetProcAddress(hMIXLib,"mix_init");
   assert(mixinit != NULL);
   return mixinit(mixdir);
}
//---------------------------------------------------------------------------

void __fastcall TMixmaster3::mix_exit()
{
   MIX_EXIT mixexit = (MIX_EXIT)GetProcAddress(hMIXLib, "mix_exit");
   assert(mixexit != NULL);
   return mixexit();
}
//---------------------------------------------------------------------------

void __fastcall TMixmaster3::rand_bytes(unsigned char *buf, int num)
{
   // TODO 0: Mixmaster needs mouse moves to rand_bytes
   RAND_BYTES randbytes = (RAND_BYTES)GetProcAddress(hSSLLib, "RAND_bytes");
   assert(randbytes != NULL);
   randbytes(buf, num);
}
//---------------------------------------------------------------------------

int __fastcall TMixmaster3::data_encrypt(char **data, long &datlen,
   char *password, int paslen)
{
   assert(data != NULL && datlen > 0 && password != NULL && paslen > 0);

   int result;
   BUFFER *dat = buf_new(), *pas = buf_new();
   dat->sensitive = true; pas->sensitive = true;
   buf_append(pas,(byte*)password,paslen);
   buf_append(dat,(byte*)*data,datlen);
   if ((result = pgp_encrypt(PGP_CONVENTIONAL|PGP_NOARMOR,dat,pas,NULL,NULL,NULL,NULL))==0)
   {
      delete *data; *data = new char[dat->length+1];
      // copy uncompressed data to data variable
      memcpy(*data,dat->data,dat->length+1);
      // set new length
      datlen = dat->length;
   }
   buf_free(pas); buf_free(dat);
   return result;
}
//---------------------------------------------------------------------------

int __fastcall TMixmaster3::data_decrypt(char** data, long& datlen,
   char *password, int paslen)
{
   assert(data != NULL && datlen > 0 && password != NULL && paslen > 0);

   int result;
   BUFFER *dat = buf_new(), *pas = buf_new();
   dat->sensitive = true; pas->sensitive = true;
   buf_append(pas,(byte*)password,paslen);
   buf_append(dat,(byte*)*data,datlen);
   if ((result = pgp_decrypt(dat,pas,NULL,NULL,NULL))==0)
   {
      delete *data; *data = new char[dat->length+1];
      // copy uncompressed data to data variable
      memcpy(*data,dat->data,dat->length+1);
      // set new length
      datlen = dat->length;
   }
   buf_free(pas); buf_free(dat);
   return result;
}
//---------------------------------------------------------------------------

int __fastcall TMixmaster3::pgp_encrypt(int mode, BUFFER *message, BUFFER *encr,
   BUFFER *sigid, BUFFER *pass, char *pubring, char *secring)
{
   PGP_ENCRYPT encrypt = (PGP_ENCRYPT)GetProcAddress(hMIXLib,"pgp_encrypt");
   assert(encrypt != NULL);
   return encrypt(mode,message,encr,sigid,pass,pubring,secring);
}
//---------------------------------------------------------------------------

int __fastcall TMixmaster3::pgp_mailenc(int mode, BUFFER *message, char *sigid,
   BUFFER *pass, char *pubring, char *secring)
{
   PGP_MAILENC mailenc = (PGP_MAILENC)GetProcAddress(hMIXLib,"pgp_mailenc");
   assert(mailenc != NULL);
   return mailenc(mode,message,sigid,pass,pubring,secring);
}
//---------------------------------------------------------------------------

int __fastcall TMixmaster3::pgp_decrypt(BUFFER *message, BUFFER *pass, BUFFER *sig,
   char *pubring, char *secring)
{
   PGP_DECRYPT decrypt = (PGP_DECRYPT)GetProcAddress(hMIXLib,"pgp_decrypt");
   assert(decrypt != NULL);
   return decrypt(message,pass,sig,pubring,secring);
}

//---------------------------------------------------------------------------
// MIX BUFFERS Copied
//---------------------------------------------------------------------------

int __fastcall buf_rewind(BUFFER *buffer)
{
  buffer->ptr = 0;
  return (0);
}
//---------------------------------------------------------------------------

int __fastcall buf_sets(BUFFER *buffer, char *s)
{
   buf_clear(buffer);
   return (buf_appends(buffer, s));
}
//---------------------------------------------------------------------------

int __fastcall buf_getc(BUFFER *buffer)
{
  if (buffer->ptr == buffer->length) return (-1);
  else return (buffer->data[buffer->ptr++]);
}
//---------------------------------------------------------------------------

long __fastcall buf_getl(BUFFER *b)
{
  long l;
  if (b->ptr + 4 > b->length) return (-1);
  l  = buf_getc(b) << 24;
  l += buf_getc(b) << 16;
  l += buf_getc(b) << 8;
  l += buf_getc(b);
  return (l);
}
//---------------------------------------------------------------------------

int __fastcall buf_geti(BUFFER *b)
{
  int i;
  if (b->ptr + 2 > b->length) return (-1);
  i  = buf_getc(b) << 8;
  i += buf_getc(b);
  return (i);
}
//---------------------------------------------------------------------------

int __fastcall buf_get(BUFFER *buffer, BUFFER *to, int n)
{
  buf_clear(to);
  if (buffer->length - buffer->ptr < n)
    return (-1);
  buf_append(to, buffer->data + buffer->ptr, n);
  buffer->ptr += n;
  return (0);
}

//---------------------------------------------------------------------------

int __fastcall buf_nl(BUFFER *b)
{
   return (buf_append(b,(byte*)"\n", 1));
}

//---------------------------------------------------------------------------
// MIX BUFFERS
//---------------------------------------------------------------------------

BUFFER* __fastcall buf_new(void)
{
   BUF_NEW bufnew = (BUF_NEW)GetProcAddress(hMIXLib,"buf_new");
   assert(bufnew != NULL);
   return bufnew();
}
//---------------------------------------------------------------------------

int __fastcall buf_free(BUFFER *buf)
{
   BUF_FREE buffree = (BUF_FREE)GetProcAddress(hMIXLib,"buf_free");
   assert(buffree != NULL);
   return buffree(buf);
}
//---------------------------------------------------------------------------

int __fastcall buf_append(BUFFER *buffer, byte *mess, int len)
{
   BUF_APPEND bufappend = (BUF_APPEND)GetProcAddress(hMIXLib,"buf_append");
   assert(bufappend != NULL);
   return bufappend(buffer,mess,len);
}
//---------------------------------------------------------------------------

int __fastcall buf_reset(BUFFER *buf)
{
   BUF_RESET bufreset = (BUF_RESET)GetProcAddress(hMIXLib, "buf_reset");
   assert(bufreset != NULL);
   return bufreset(buf);
}
//---------------------------------------------------------------------------

int __fastcall buf_appends(BUFFER *buffer, char *s)
{
   BUF_APPENDS bufappends = (BUF_APPENDS)GetProcAddress(hMIXLib, "buf_appends");
   assert(bufappends != NULL);
   return bufappends(buffer, s);
}
//---------------------------------------------------------------------------

int __fastcall buf_clear(BUFFER *buffer)
{
   BUF_CLEAR bufclear = (BUF_CLEAR)GetProcAddress(hMIXLib, "buf_clear");
   assert(bufclear != NULL);
   return bufclear(buffer);
}
//---------------------------------------------------------------------------

int buf_readfile(BUFFER *buffer, char *filename)
{
   int err = -1;
   buf_reset(buffer);
   TFileStream *f;
   if (!FileExists(filename))
   {
      ShowMessage(
"The file "+AnsiString(filename)+" does not exist! Unable to complete operation."
      );
      return err;
   }
   try
   {
      f = new TFileStream(filename,fmOpenRead);
   }
   catch (...)
   {
      ShowMessage(
"I was unable to open "+AnsiString(filename)+" !\n\
 Please make sure it is not in use by another program, and try\n\
 the operation again."
      );
      f = NULL;
   }
   if (f)
   {
      char buf[BUFSIZE] = { 0 };
      while (true)
      {
         if (!f->Size) { err = 0; break; }

         int n = f->Read(buf,BUFSIZE); if (n > 0) err = 0;
         buf_append(buffer, (byte*)buf, n);
         if (n < BUFSIZE) break;

         if (buffer->length > BUFFER_MAX)
         {
            ShowMessage("Message file too large. Giving up.");
            return(1);
         }
      }
      delete f;
   }
   return err;
}
//---------------------------------------------------------------------------

int __fastcall buf_appendl(BUFFER *to, long l)
{
   BUF_APPENDL bufappendl = (BUF_APPENDL)GetProcAddress(hMIXLib, "buf_appendl");
   assert(bufappendl != NULL);
   return bufappendl(to, l);
}
//---------------------------------------------------------------------------

int __fastcall buf_appendc(BUFFER *to, byte b)
{
   BUF_APPENDC bufappendc = (BUF_APPENDC)GetProcAddress(hMIXLib, "buf_appendc");
   assert(bufappendc != NULL);
   return bufappendc(to, b);
}
//---------------------------------------------------------------------------

int __fastcall buf_rest(BUFFER *to, BUFFER *from)
{
   BUF_REST bufrest = (BUF_REST)GetProcAddress(hMIXLib, "buf_rest");
   assert(bufrest != NULL);
   return bufrest(to,from);
}
//---------------------------------------------------------------------------

int __fastcall buf_appendi(BUFFER *to, int i)
{
   BUF_APPENDI bufappendi = (BUF_APPENDI)GetProcAddress(hMIXLib, "buf_appendi");
   assert(bufappendi != NULL);
   return bufappendi(to, i);
}
//---------------------------------------------------------------------------

int __fastcall buf_getline(BUFFER *buffer, BUFFER *line)
{
   BUF_GETLINE bufgetline = (BUF_GETLINE)GetProcAddress(hMIXLib, "buf_getline");
   assert(bufgetline != NULL);
   return bufgetline(buffer, line);
}
//---------------------------------------------------------------------------

int __fastcall buf_getheader(BUFFER *buffer, BUFFER *field, BUFFER *content)
{
   BUF_GETHEADER bufgetheader = (BUF_GETHEADER)GetProcAddress(hMIXLib, "buf_getheader");
   assert(bufgetheader != NULL);
   return bufgetheader(buffer, field, content);
}
//---------------------------------------------------------------------------

int __fastcall decode(BUFFER *in, BUFFER *out)
{
   DECODE decode = (DECODE)GetProcAddress(hMIXLib, "decode");
   assert(decode != 0);
   return decode(in, out);
}
//---------------------------------------------------------------------------

int __fastcall encode(BUFFER *in, int linelen)
{
   ENCODE encode = (ENCODE)GetProcAddress(hMIXLib, "encode");
   assert(encode != NULL);
   return encode(in, linelen);
}
//---------------------------------------------------------------------------
